import json

import pytest
from jmespath.exceptions import JMESPathTypeError

from dql.jmes_sql.transpiler import Transpiler

# pylint: disable=unused-argument,redefined-outer-name

cat1 = {
    "class": "cat",
    "id": "1",
    "num_annotators": 5,
    "inference": {"class": "cat", "confidence": 0.45},
}

cat10 = {
    "class": "cat",
    "id": "10",
    "num_annotators": 8,
    "inference": {"class": "dog", "confidence": 0.86},
}

open_image1 = {
    "classifications": [
        {
            "Confidence": 1,
            "LabelName": "/m/0b_rs",
            "Source": "verification",
            "some_score": 0.72,
        },
        {
            "Confidence": 1,
            "LabelName": "/m/0bt9lr",
            "Source": "crowdsource-verification",
            "some_score": 0.38,
        },
        {
            "Confidence": 0,
            "LabelName": "/m/015p6",
            "Source": "verification",
            "some_score": 0.934,
        },
        {
            "Confidence": 1,
            "LabelName": "/m/01lrl",
            "Source": "verification",
            "some_score": 0.394,
        },
        {
            "Confidence": 0,
            "LabelName": "/m/02_n6y",
            "Source": "verification",
            "some_score": 0.38,
        },
    ],
    "detections": [
        {
            "Confidence": 1,
            "IsDepiction": 0,
            "IsGroupOf": 0,
            "IsInside": 0,
            "IsOccluded": 1,
            "IsTruncated": 1,
            "LabelName": "/m/0b_rs",
            "Source": "xclick",
            "XClick1X": 0.0,
            "XClick1Y": 0.998333,
            "XClick2X": 0.99875,
            "XClick2Y": 0.0,
            "XClick3X": 0.99875,
            "XClick3Y": 0.0,
            "XClick4X": 0.99875,
            "XClick4Y": 0.0,
            "XMax": 0.99875,
            "XMin": 0.0,
            "YMax": 0.998333,
            "YMin": 0.0,
        },
        {
            "Confidence": 1,
            "IsDepiction": 0,
            "IsGroupOf": 0,
            "IsInside": 0,
            "IsOccluded": 1,
            "IsTruncated": 1,
            "LabelName": "/m/0bt9lr",
            "Source": "xclick",
            "XClick1X": 0.63625,
            "XClick1Y": 0.268333,
            "XClick2X": 0.90375,
            "XClick2Y": 0.506667,
            "XClick3X": 0.74875,
            "XClick3Y": 0.998333,
            "XClick4X": 0.165,
            "XClick4Y": 0.661667,
            "XMax": 0.90375,
            "XMin": 0.165,
            "YMax": 0.998333,
            "YMin": 0.268333,
        },
    ],
    "id": "0000b9fcba019d36",
    "image_ids": {
        "Author": "O. G.",
        "AuthorProfileURL": "https://www.flickr.com/people/ezlens/",
        "License": "https://creativecommons.org/licenses/by/2.0/",
        "OriginalLandingURL": "https://www.flickr.com/photos/ezlens/50621395",
        "OriginalMD5": "3d9qBaSrX7fN2mag/qpk5w==",
        "OriginalSize": 73392,
        "OriginalURL": "https://farm7.staticflickr.com/29/50621395_874fb9f3ee_o.jpg",  # noqa: E501
        "Rotation": None,
        "Subset": "train",
        "Thumbnail300KURL": "https://c7.staticflickr.com/1/29/50621395_2c80281034_z.jpg?zz=1",  # noqa: E501
        "Title": "DSCN8023",
    },
    "segmentations": [
        {
            "BoxID": "99ba1430",
            "BoxXMax": 0.90375,
            "BoxXMin": 0.165,
            "BoxYMax": 0.998333,
            "BoxYMin": 0.268333,
            "Clicks": "0.64587 0.27696 1;0.36091 0.83590 0;0.76364 0.98524 1;0.85670 0.65910 0;0.31044 0.85548 1;0.69804 0.29354 1;0.76067 0.38381 0;0.87069 0.60723 1;0.70049 0.83122 1;0.75876 0.32617 1;0.49428 0.61008 1",  # noqa: E501
            "LabelName": "/m/0bt9lr",
            "MaskPath": "0000b9fcba019d36_m0bt9lr_99ba1430.png",
            "PredictedIoU": 0.74734,
        }
    ],
    "split": "train",
}


@pytest.fixture
def con(data_storage):
    conn = data_storage.db  # pylint:disable=protected-access

    # These tests re-use the same database for better performance
    result = conn.execute(
        """
        SELECT name FROM sqlite_master
        WHERE type='table' AND name='meta'
        """
    ).fetchall()

    if len(result):
        return conn

    conn.execute(
        """
        CREATE TABLE meta (
            name    VARCHAR,
            data    JSON
        )
        """
    )

    conn.execute(
        "INSERT INTO meta VALUES (?, ?), (?, ?), (?, ?)",
        [
            "cat1",
            json.dumps(cat1),
            "cat10",
            json.dumps(cat10),
            "oi_dog",
            json.dumps(open_image1),
        ],
    )
    conn.commit()
    return conn


def test_basic(con):
    res = con.execute("SELECT COUNT() FROM meta").fetchall()[0][0]
    assert res == 3


def test_confident_cat(con):
    res = con.execute(
        """
        SELECT name
        FROM meta
        WHERE data -> 'inference' ->> 'confidence' > 0.5
    """
    ).fetchall()

    assert len(res) == 1
    assert res[0][0] == "cat10"


def test_func_contains_string_match(con):
    query = "contains(`foobar`, `foo`)"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()
    assert len(res) == 3


def test_func_contains_string_missmatch(con):
    query = "contains(`foobar`, `omg`)"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()
    assert len(res) == 0


def test_func_contains_string_error(con):
    query = "contains(`false`, `omg`)"
    with pytest.raises(JMESPathTypeError):
        ts = Transpiler("meta", "data", ["name", "data"])
        ts.translate(query)


def test_func_contains_field(con):
    query = "contains(class, `at`)"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()

    assert len(res) == 2
    assert {res[0][0], res[1][0]} == {"cat1", "cat10"}


def test_func_contains_field_subexpression(con):
    query = "contains(inference.class, `at`)"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()

    assert len(res) == 1
    assert res[0][0] == "cat1"


def test_subexpression_equals_to_string(con):
    query = "inference.class == `dog`"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()

    assert len(res) == 1
    assert res[0][0] == "cat10"


def test_subexpression_equals_to_subexpression(con):
    query = "class == inference.class"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()

    assert len(res) == 1
    assert res[0][0] == "cat1"


def test_larger_than_a_number(con):
    query = "num_annotators > `6`"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()

    assert len(res) == 1
    assert res[0][0] == "cat10"


def test_not_sql_number_larger(con):
    query = "`872` > `539`"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()
    assert len(res) == 3


def test_not_sql_number_equal(con):
    query = "`4546` == `4546`"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()
    assert len(res) == 3


def test_not_sql_number_larger_error(con):
    query = "`34` > `834`"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()
    assert len(res) == 0


def test_func_contains_array(con):
    query = "detections[1].LabelName == `/m/0bt9lr`"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()

    assert len(res) == 1
    assert res[0][0] == "oi_dog"


def test_func_contains_array_error(con):
    query = "detections[1].LabelName == `/m/0b_rs`"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()
    assert len(res) == 0


def test_jmes_contains_with_projection(con):
    # -- DuckDB solution:
    # SELECT
    #     list_contains(
    #         list_transform(from_json(data -> > 'detections', '["json"]'),
    #                 x -> json_extract_string(json(x), 'LabelName') ),
    #         '/m/0bt9lr')
    # FROM meta

    query = "contains(detections[*].LabelName, `/m/0bt9lr`)"
    cursor = Transpiler.execute_query(con, query, "meta", "data", ["name", "data"])
    res = cursor.fetchall()

    assert len(res) == 1
    assert res[0][0] == "oi_dog"


def test_two_tables():
    # TODO: complete test
    # query = (
    #    "contains(detections[*].LabelName,"
    #    "max_by(classifications, &some_score).LabelName)"
    # )
    pass



import io

import pytest
from hypothesis import HealthCheck, given, settings
from hypothesis import strategies as st

from dql.client import Client
from dql.client.fileslice import FileSlice

from ..utils import uppercase_scheme


def test_bad_url():
    bucket = "whatever"
    path = "my/path"
    with pytest.raises(RuntimeError):
        Client.parse_url(bucket + "/" + path + "/", None)


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(rel_path=st.text(min_size=1))
def test_parse_url(cloud_test_catalog, rel_path):
    bucket_uri = cloud_test_catalog.src_uri
    url = f"{bucket_uri}/{rel_path}"
    catalog = cloud_test_catalog.catalog
    client, rel_part = catalog.parse_url(url, **cloud_test_catalog.client_config)
    assert client.uri == bucket_uri
    assert rel_part == rel_path


@settings(suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(rel_path=st.text(min_size=1))
def test_parse_url_uppercase_scheme(cloud_test_catalog, rel_path):
    bucket_uri = cloud_test_catalog.src_uri
    bucket_uri_upper = uppercase_scheme(bucket_uri)
    url = f"{bucket_uri_upper}/{rel_path}"
    catalog = cloud_test_catalog.catalog
    client, rel_part = catalog.parse_url(url, **cloud_test_catalog.client_config)
    assert client.uri == bucket_uri
    assert rel_part == rel_path


def test_FileSlice():
    data = b"0123456789abcdef"
    base = io.BytesIO(data)
    f = FileSlice(base, 5, 5, "foo")
    assert base.tell() == 0
    assert f.readable()
    assert not f.writable()
    assert f.seekable()
    assert f.name == "foo"

    # f.seek() doesn't move the underlying stream
    f.seek(0)
    assert f.tell() == 0
    assert base.tell() == 0

    assert f.read(3) == data[5:8]
    assert f.tell() == 3
    assert base.tell() == 8

    assert f.read(4) == data[8:10]
    assert f.tell() == 5
    assert base.tell() == 10

    b = bytearray(5)
    f.seek(0)
    f.readinto(b)
    assert b == data[5:10]


def test_bad_FileSlice():
    data = b"0123456789abcdef"
    base = io.BytesIO(data)
    f = FileSlice(base, 10, 10, "foo")
    assert f.read(4) == data[10:14]
    with pytest.raises(RuntimeError):
        f.read()



import logging
from argparse import ArgumentTypeError

import pytest

from dql.cli import (
    TTL_HUMAN,
    TTL_INT,
    find_columns_type,
    get_logging_level,
    get_parser,
    human_time_type,
)


def test_human_time_type():
    assert human_time_type(TTL_HUMAN) == TTL_INT
    assert human_time_type("1h") == 60 * 60
    assert human_time_type("30s") == 30
    assert human_time_type("2w") == 2 * 7 * 24 * 60 * 60

    assert human_time_type("", can_be_none=True) is None

    with pytest.raises(ArgumentTypeError):
        human_time_type("bogus")


def test_find_columns_type():
    assert find_columns_type("") == ["path"]
    assert find_columns_type("du") == ["du"]
    assert find_columns_type("", default_colums_str="name") == ["name"]
    assert find_columns_type("du, name,PATH") == ["du", "name", "path"]

    with pytest.raises(ArgumentTypeError):
        find_columns_type("bogus")


def test_cli_parser():
    parser = get_parser()

    args = parser.parse_args(("ls", "s3://example-bucket/", "--ttl", "1d"))

    assert args.ttl == 24 * 60 * 60
    assert args.sources == ["s3://example-bucket/"]

    assert args.quiet == 0
    assert args.verbose == 0

    assert get_logging_level(args) == logging.INFO

    args = parser.parse_args(("ls", "s3://example-bucket/", "-vvv"))

    assert args.quiet == 0
    assert args.verbose == 3

    assert get_logging_level(args) == logging.DEBUG

    args = parser.parse_args(("ls", "s3://example-bucket/", "-q"))

    assert args.quiet == 1
    assert args.verbose == 0

    assert get_logging_level(args) == logging.CRITICAL



import json
import logging
import os

from dql.client.checksum import md5
from dql.nodes_thread_pool import NodesThreadPool

logger = logging.getLogger("dql")


class NodesFetcher(NodesThreadPool):
    def __init__(self, client, data_storage, file_path, max_threads, cache):
        super().__init__(max_threads)
        self.client = client
        self.data_storage = data_storage
        self.file_path = file_path
        self.cache = cache

    def done_task(self, done):
        updated_nodes = []
        for d in done:
            lst = d.result()
            for node, checksum in lst:
                self.data_storage.update_checksum(node, checksum)
                node.checksum = checksum
                updated_nodes.append(node)
        return updated_nodes

    def do_task(self, chunk):
        res = []
        for node in chunk:
            if self.cache.exists(node.checksum):
                self.increase_counter(node.size)
                continue

            pair = self.fetch(self.client.name, node.path_str, node)
            res.append(pair)
        return res

    def fetch(self, bucket, path, node):
        from dvc_data.hashfile.build import _upload_file
        from dvc_objects.fs.callbacks import Callback

        if node.vtype == "tar":
            return self._fetch_from_tar(node)

        class _CB(Callback):
            def relative_update(  # pylint: disable=no-self-argument
                _, inc: int = 1  # noqa: disable=no-self-argument
            ):
                self.increase_counter(inc)

        _, obj = _upload_file(
            f"{bucket}/{path}",
            self.client.fs,
            self.cache,
            self.cache,
            callback=_CB(),
        )

        return node, obj.hash_info.value

    def _fetch_from_tar(self, node):
        assert node.location is not None
        loc_stack = json.loads(node.location)
        if len(loc_stack) > 1:
            raise NotImplementedError("Nested v-objects are not supported yet.")
        location = loc_stack[0]
        offset = location["offset"]
        tar_path = location["parent"]
        with self.client.open(tar_path) as f:
            f.seek(offset)
            contents = f.read(node.size)
        checksum = md5(contents).hexdigest()
        dst = self.cache.oid_to_path(checksum)
        if not os.path.exists(dst):
            # Create the file only if it's not already in cache
            os.makedirs(os.path.dirname(dst), exist_ok=True)
            with open(dst, mode="wb") as f:
                f.write(contents)
        self.increase_counter(node.size)
        return node, checksum



class DQLError(RuntimeError):
    pass


class NotFoundError(Exception):
    pass


class DatasetNotFoundError(NotFoundError):
    pass


class StorageNotFoundError(NotFoundError):
    pass


class PendingIndexingError(Exception):
    """An indexing operation is already in progress."""



from .catalog import Catalog, parse_edql_file
from .formats import indexer_formats
from .loader import get_catalog

__all__ = [
    "Catalog",
    "get_catalog",
    "indexer_formats",
    "parse_edql_file",
]



import logging
import os
import posixpath
import sqlite3
from datetime import MAXYEAR, MINYEAR, datetime, timezone
from functools import wraps
from time import sleep
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import sqlalchemy
from sqlalchemy import MetaData, Table, UniqueConstraint, select, update
from sqlalchemy.dialects import sqlite
from sqlalchemy.schema import Column as saColumn
from sqlalchemy.schema import CreateTable, DropTable
from sqlalchemy.sql import func
from sqlalchemy.sql.expression import bindparam, cast, true

from dql.data_storage.abstract import AbstractDataStorage
from dql.data_storage.schema import DATASET_CORE_COLUMN_NAMES, DefaultSchema
from dql.dataset import DatasetRecord
from dql.dataset import Status as DatasetStatus
from dql.error import DQLError
from dql.storage import Status as StorageStatus
from dql.storage import Storage
from dql.utils import DQLDir

if TYPE_CHECKING:
    from sqlalchemy.schema import SchemaItem
    from sqlalchemy.sql.elements import ColumnClause, CompilerElement, TextClause
    from sqlalchemy.types import TypeEngine

logger = logging.getLogger("dql")

RETRY_START_SEC = 0.01
RETRY_MAX_TIMES = 10
RETRY_FACTOR = 2

Column = Union[str, "ColumnClause[Any]", "TextClause"]

# sqlite 3.31.1 is the earliest version tested in CI
if sqlite3.sqlite_version_info < (3, 31, 1):
    logger.warning(
        "Possible sqlite incompatibility. The earliest tested version of "
        f"sqlite is 3.31.1 but you have {sqlite3.sqlite_version}"
    )


sqlite_dialect = sqlite.dialect(paramstyle="named")
quote_schema = sqlite_dialect.identifier_preparer.quote_schema
quote = sqlite_dialect.identifier_preparer.quote


def compile_statement(
    statement: "CompilerElement",
) -> Union[Tuple[str], Tuple[str, Dict[str, Any]]]:
    # render_postcompile is needed for in_ queries to work
    compiled = statement.compile(
        dialect=sqlite_dialect, compile_kwargs={"render_postcompile": True}
    )
    if compiled.params is None:
        return (compiled.string,)
    return compiled.string, compiled.params


def adapt_datetime(val: datetime) -> str:
    if not (val.tzinfo is timezone.utc or val.tzname() == "UTC"):
        try:
            val = val.astimezone(timezone.utc)
        except (OverflowError, ValueError, OSError):
            if val.year == MAXYEAR:
                val = datetime.max
            elif val.year == MINYEAR:
                val = datetime.min
            else:
                raise
    return val.replace(tzinfo=None).isoformat(" ")


def convert_datetime(val: bytes) -> datetime:
    return datetime.fromisoformat(val.decode()).replace(tzinfo=timezone.utc)


sqlite3.register_adapter(datetime, adapt_datetime)
sqlite3.register_converter("datetime", convert_datetime)


def get_retry_sleep_sec(retry_count: int) -> int:
    return RETRY_START_SEC * (RETRY_FACTOR**retry_count)


def retry_sqlite_locks(func):  # pylint: disable=redefined-outer-name
    # This retries the database modification in case of concurrent access
    @wraps(func)
    def wrapper(*args, **kwargs):
        exc = None
        for retry_count in range(RETRY_MAX_TIMES):
            try:
                return func(*args, **kwargs)
            except sqlite3.OperationalError as operror:
                exc = operror
                sleep(get_retry_sleep_sec(retry_count))
        raise exc

    return wrapper


class SQLiteDataStorage(AbstractDataStorage):
    """
    SQLite data storage uses SQLite3 for storing indexed data locally.
    This is currently used for the local cli.
    """

    def __init__(self, db_file: Optional[str] = None, uri: str = ""):
        self.schema: "DefaultSchema" = DefaultSchema()
        super().__init__(uri)
        self.db_file = db_file if db_file else DQLDir.find().db
        detect_types = sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES

        try:
            if self.db_file == ":memory:":
                # Enable multithreaded usage of the same in-memory db
                self.db = sqlite3.connect(
                    "file::memory:?cache=shared", uri=True, detect_types=detect_types
                )
            else:
                self.db = sqlite3.connect(self.db_file, detect_types=detect_types)
            self.engine = sqlalchemy.create_engine(
                "sqlite+pysqlite:///", creator=lambda: self.db, future=True
            )

            self.db.isolation_level = None  # Use autocommit mode
            self.db.execute("PRAGMA foreign_keys = ON")
            self.db.execute("PRAGMA cache_size = -102400")  # 100 MiB
            # Enable Write-Ahead Log Journaling
            self.db.execute("PRAGMA journal_mode = WAL")
            self.db.execute("PRAGMA synchronous = NORMAL")
            self.db.execute("PRAGMA case_sensitive_like = ON")
            if os.environ.get("DEBUG_SHOW_SQL_QUERIES"):
                self.db.set_trace_callback(print)

            self._init_storage_table()
            self._init_datasets_tables()
        except RuntimeError:
            raise DQLError("Can't connect to SQLite DB")

    @staticmethod
    def buckets_constraints() -> List["SchemaItem"]:
        return [
            UniqueConstraint("uri"),
        ]

    @staticmethod
    def datasets_constraints() -> List["SchemaItem"]:
        return [
            UniqueConstraint("name"),
        ]

    @staticmethod
    def buckets_columns() -> List["SchemaItem"]:
        columns = super(SQLiteDataStorage, SQLiteDataStorage).buckets_columns()
        return columns + SQLiteDataStorage.buckets_constraints()

    @staticmethod
    def datasets_columns() -> List["SchemaItem"]:
        columns = super(SQLiteDataStorage, SQLiteDataStorage).datasets_columns()
        return columns + SQLiteDataStorage.datasets_constraints()

    def _init_storage_table(self):
        """Initialize only tables related to storage, e.g s3"""
        self.execute(CreateTable(self.storages, if_not_exists=True))

    def _init_datasets_tables(self) -> None:
        self.execute(CreateTable(self.datasets, if_not_exists=True))
        self.execute(CreateTable(self.datasets_versions, if_not_exists=True))

    def init_db(self, prefix: str = "", is_new: bool = True):
        assert self.listing_table_name, "Missing listing_table_name"
        table = self.nodes.table
        partials_table = self.partials
        ig = self.id_generator
        if not prefix or is_new:
            self.execute(DropTable(table, if_exists=True))
            self.execute(DropTable(partials_table, if_exists=True))
            # Do not drop ID Generator, as this is shared across all buckets
        self.execute(CreateTable(table, if_not_exists=True))
        self.execute(CreateTable(partials_table, if_not_exists=True))
        self.execute(CreateTable(ig, if_not_exists=True))
        if not prefix or is_new:
            # If there is no prefix (indexing the whole bucket) or this is a new
            # bucket entry, then reset the id_generator for this bucket by deleting
            # any existing used last_id entries, and reinsert the zero last_id
            # as the start, as that indicates no ids have been used yet.
            partials_uri = f"partials:{self.uri}"
            self.execute(
                self.id_generator_delete().where(
                    (ig.c.uri == self.uri) | (ig.c.uri == partials_uri)
                )
            )
            self.execute(self.id_generator.insert().values(uri=self.uri, last_id=0))
            self.execute(self.id_generator.insert().values(uri=partials_uri, last_id=0))

    def clone(self, uri: Optional[str] = None) -> "SQLiteDataStorage":
        uri = uri or self.uri
        return SQLiteDataStorage(db_file=self.db_file, uri=uri)

    @retry_sqlite_locks
    def execute(
        self,
        query,
        cursor: Optional[sqlite3.Cursor] = None,
        conn=None,
    ) -> sqlite3.Cursor:
        if cursor:
            return cursor.execute(*compile_statement(query))
        return self.db.execute(*compile_statement(query))

    @retry_sqlite_locks
    def executemany(
        self, query, params, cursor: Optional[sqlite3.Cursor] = None
    ) -> sqlite3.Cursor:
        if cursor:
            return cursor.executemany(compile_statement(query)[0], params)
        return self.db.executemany(compile_statement(query)[0], params)

    def _get_next_ids(self, uri: str, count: int) -> range:
        with self.db:
            # Transactions ensure no concurrency conflicts
            self.db.execute("BEGIN")
            ig = self.id_generator
            self.execute(
                self.id_generator_update()
                .where(ig.c.uri == uri)
                .values(last_id=ig.c.last_id + count)
            )
            # TODO: RETURNING might be a better option,
            # but is only supported on SQLite 3.35.0 or newer
            last_id = self.execute(
                self.id_generator_select(ig.c.last_id).where(ig.c.uri == uri)
            ).fetchone()[0]

        return range(last_id - count + 1, last_id + 1)

    def create_dataset_rows_table(
        self,
        name: str,
        custom_columns: Sequence["sqlalchemy.Column"] = (),
        if_not_exists: bool = True,
    ) -> None:
        q = CreateTable(
            self.schema.dataset_row_cls.new_table(
                name, custom_columns=custom_columns, metadata=self.metadata
            ),
            if_not_exists=if_not_exists,
        )
        self.execute(q)

    def _get_dataset_row_values(
        self,
        name: str,
        columns: Optional[Sequence[str]] = None,
        limit: Optional[int] = 20,
        version=None,
    ) -> Iterator[Mapping[str, Any]]:
        dataset = self.get_dataset(name)

        dr = self.dataset_rows(dataset.id, version)
        select_columns = []
        if columns:
            select_columns = [getattr(dr.c, c) for c in columns]
        query = dr.select(*select_columns)
        if limit:
            query = query.limit(limit)

        cur = self.db.cursor()
        cur.row_factory = sqlite3.Row  # type: ignore[assignment]
        yield from self.execute(query, cursor=cur)

    def create_shadow_dataset(
        self,
        name: str,
        create_rows: Optional[bool] = True,
        custom_columns: Sequence["sqlalchemy.Column"] = (),
    ) -> DatasetRecord:
        """Creates new shadow dataset if it doesn't exist yet"""
        with self.db:
            self.db.execute("begin")

            d = self.datasets
            self.execute(
                sqlite.insert(d)
                .values(
                    name=name,
                    shadow=True,
                    status=DatasetStatus.CREATED,
                    created_at=datetime.now(timezone.utc),
                )
                .on_conflict_do_nothing(index_elements=["name"])
            )
            dataset = self.get_dataset(name)
            assert dataset.shadow

            # Reassign custom columns to the new table.
            custom_cols = [
                saColumn(c.name, c.type, nullable=c.nullable) for c in custom_columns
            ]
            if create_rows:
                table_name = self.dataset_table_name(dataset.id)
                self.create_dataset_rows_table(table_name, custom_cols)
                dataset = self.update_dataset_status(dataset, DatasetStatus.PENDING)

            return dataset  # type: ignore[return-value]

    def insert_into_shadow_dataset(
        self, name: str, uri: str, path: str, recursive=False
    ) -> None:
        dataset = self.get_dataset(name)
        assert dataset.shadow

        dst = self.dataset_rows(dataset.id)  # Destination (dataset table).
        src = self.nodes_table(uri)  # Source (nodes table).

        src_cols = {c.name for c in src.c}
        # Not including id
        dst_cols = {c.name for c in dst.c if c.name != "id"}
        cols = src_cols.intersection(dst_cols)
        cols.add("source")
        transfer_fields = list(cols)
        select_query = self.nodes_dataset_query(
            transfer_fields, path=path, recursive=recursive, uri=uri
        )

        insert_query = dst.insert().from_select(transfer_fields, select_query)
        self.execute(insert_query)

    def _rename_table(self, old_name: str, new_name: str):
        comp_old_name = quote_schema(old_name)
        comp_new_name = quote_schema(new_name)
        self.db.execute(f"ALTER TABLE {comp_old_name} RENAME TO {comp_new_name}")

    def create_dataset_version(
        self, name: str, version: int, create_rows_table=True
    ) -> DatasetRecord:
        with self.db:
            self.db.execute("begin")

            dataset = self.get_dataset(name)

            dv = self.datasets_versions
            self.execute(
                sqlite.insert(dv)
                .values(
                    dataset_id=dataset.id,
                    version=version,
                    created_at=datetime.now(timezone.utc),
                )
                .on_conflict_do_nothing(index_elements=["dataset_id", "version"])
            )

            if create_rows_table:
                table_name = self.dataset_table_name(dataset.id, version)
                self.create_dataset_rows_table(table_name)

            return dataset

    def merge_dataset_rows(
        self,
        src: DatasetRecord,
        dst: DatasetRecord,
        src_version: Optional[int] = None,
        dst_version: Optional[int] = None,
    ) -> None:
        dst_empty = False

        if not self.has_table(self.dataset_table_name(src.id, src_version)):
            # source table doesn't exist, nothing to do
            return

        if not self.has_table(self.dataset_table_name(dst.id, dst_version)):
            # destination table doesn't exist, create it
            self.create_dataset_rows_table(self.dataset_table_name(dst.id))
            dst_empty = True

        src_dr = self.dataset_rows(src.id, src_version).table
        dst_dr = self.dataset_rows(dst.id, dst_version).table
        dst_dr_latest = self.dataset_rows(dst.id, dst.latest_version).table

        # Not including id
        merge_fields = [c for c in DATASET_CORE_COLUMN_NAMES if c != "id"]

        select_src = select(*(getattr(src_dr.c, f) for f in merge_fields))
        select_dst_latest = select(*(getattr(dst_dr_latest.c, f) for f in merge_fields))

        union_query = sqlalchemy.union(select_src, select_dst_latest)

        if dst_empty:
            # we don't need union, but just select from source to destination
            insert_query = sqlite.insert(dst_dr).from_select(merge_fields, select_src)
        else:
            insert_query = (
                sqlite.insert(dst_dr)
                .from_select(merge_fields, union_query)
                .prefix_with("OR IGNORE")
            )

        self.execute(insert_query)

    def remove_shadow_dataset(self, dataset: DatasetRecord, drop_rows=True) -> None:
        with self.db:
            self.db.execute("begin")
            d = self.datasets
            self.execute(self.datasets_delete().where(d.c.id == dataset.id))
            if drop_rows:
                table_name = self.dataset_table_name(dataset.id)
                self.execute(DropTable(Table(table_name, MetaData())))

    async def insert_entry(self, entry: Dict[str, Any]) -> int:
        return (
            self.execute(
                self.nodes.insert().values(self._prepare_node(entry))
            ).lastrowid
            or 0
        )

    async def insert_entries(self, entries: Iterable[Dict[str, Any]]) -> None:
        self.executemany(
            self.nodes.insert().values({f: bindparam(f) for f in self.node_fields[1:]}),
            map(self._prepare_node, entries),
        )

    def add_bucket_signal_column(self, uri: str, col_name: str, col_type: "TypeEngine"):
        """
        Adds a signal column to a bucket.
        """
        nodes = self.nodes_table(uri)
        comp_table_name = quote_schema(nodes.name)
        comp_c_type = col_type.compile(dialect=sqlite_dialect)
        comp_c_name = quote(col_name)
        q = f"ALTER TABLE {comp_table_name} ADD COLUMN {comp_c_name} {comp_c_type}"
        self.db.execute(q)

        # reload the table to self.metadata so the table object
        # self.metadata.tables[nodes.name] includes the new column
        Table(
            nodes.name, self.metadata, extend_existing=True, autoload_with=self.engine
        )
        # invalidate any nodes table cached for this uri
        if uri == self.uri:
            self._nodes = None

    def create_storage_if_not_registered(
        self, uri: str, symlinks: bool = False
    ) -> None:
        s = self.storages
        self.execute(
            sqlite.insert(s)
            .values(uri=uri, status=StorageStatus.CREATED, symlinks=symlinks)
            .on_conflict_do_nothing()
        )

    def find_stale_storages(self):
        """
        Finds all pending storages for which the last inserted node has happened
        before STALE_HOURS_LIMIT hours, and marks it as STALE
        """
        with self.db:
            self.db.execute("begin")
            s = self.storages
            pending_storages = map(
                Storage._make,
                self.execute(
                    self.storages_select().where(s.c.status == StorageStatus.PENDING)
                ),
            )
            for storage in pending_storages:
                if storage.is_stale:
                    print(f"Marking storage {storage.uri} as stale")
                    self._mark_storage_stale(storage.id)

    def register_storage_for_indexing(
        self,
        uri: str,
        force_update: bool = True,
        prefix: str = "",
    ) -> Tuple[Storage, bool, bool, bool]:
        """
        Prepares storage for indexing operation.
        This method should be called before index operation is started
        It returns:
            - storage, prepared for indexing
            - boolean saying if indexing is needed
            - boolean saying if indexing is currently pending (running)
            - boolean saying if the storage is newly registered
        """
        # This ensures that all calls to the DB are in a single transaction
        # and commit is automatically called once this function returns
        with self.db:
            self.db.execute("begin")

            # Create storage if it doesn't exist
            self.create_storage_if_not_registered(uri)
            storage = self.get_storage(uri)

            if storage.status == StorageStatus.PENDING:
                return storage, False, True, False

            elif storage.is_expired or storage.status == StorageStatus.STALE:
                storage = self.mark_storage_pending(storage)
                return storage, True, False, False

            elif storage.status == StorageStatus.COMPLETE and not force_update:
                return storage, False, False, False

            elif (
                storage.status == StorageStatus.PARTIAL and prefix and not force_update
            ):
                if self._check_partial_index_valid(prefix):
                    return storage, False, False, False
                self._delete_partial_index(prefix)
                return storage, True, False, False

            else:
                is_new = storage.status == StorageStatus.CREATED
                storage = self.mark_storage_pending(storage)
                return storage, True, False, is_new

    def mark_storage_indexed(
        self,
        uri: str,
        status: int,
        ttl: int,
        end_time: Optional[datetime] = None,
        prefix: str = "",
        partial_id: int = 0,
    ) -> None:
        if status == StorageStatus.PARTIAL and not prefix:
            raise AssertionError("Partial indexing requires a prefix")

        if end_time is None:
            end_time = datetime.now(timezone.utc)
        expires = Storage.get_expiration_time(end_time, ttl)

        with self.db:
            self.db.execute("BEGIN")

            s = self.storages
            self.execute(
                self.storages_update()
                .where(s.c.uri == uri)
                .values(  # type: ignore [attr-defined]
                    timestamp=end_time,
                    expires=expires,
                    status=status,
                    last_inserted_at=end_time,
                )
            )

            if not self.listing_table_name:
                # This only occurs in tests
                return

            p = self.partials
            if status in {StorageStatus.COMPLETE, StorageStatus.FAILED}:
                # Delete remaining partial index paths
                self.execute(self.partials_delete())
            elif status == StorageStatus.PARTIAL:
                dirprefix = posixpath.join(prefix, "")
                self.execute(
                    sqlite.insert(p)
                    .values(
                        path_str=dirprefix,
                        timestamp=end_time,
                        expires=expires,
                        partial_id=partial_id,
                    )
                    .on_conflict_do_update(
                        index_elements=["path_str"],
                        set_={"timestamp": end_time, "expires": expires},
                    )
                )

    def validate_paths(self) -> int:
        """Find and mark any invalid paths."""
        t1 = self.nodes.table
        t2 = t1.alias("t2")
        t3 = t1.alias("t3")
        id_query = (
            select(t2.c.id)  # type: ignore[attr-defined]
            .select_from(t2)
            .join(t3, (t2.c.path_str == t3.c.path_str) & (t2.c.id != t3.c.id))
            .where(
                t2.c.valid == true(),
                t2.c.dir_type == 0,
            )
        )
        query = (
            update(t1)
            .values(valid=False)
            .where(t1.c.id.in_(id_query))  # type: ignore[attr-defined]
        )

        row_count = self.execute(query).rowcount
        if row_count:
            logger.warning(
                "File names that collide with directory names will be ignored. "
                f"Number found: {row_count}"
            )
        return row_count

    def instr(self, source, target) -> sqlalchemy.Boolean:
        return cast(func.instr(source, target), sqlalchemy.Boolean)

    def get_table(self, name: str) -> sqlalchemy.Table:
        return sqlalchemy.Table(
            name,
            self.metadata,
            autoload_with=self.engine,
            extend_existing=True,
        )



from sqlalchemy import Integer

from dql.dataset import DatasetRow
from dql.query import C, generator, udf


def test_udf():
    @udf(Integer, C.id, C.size)
    def t(a, b):
        return a * b

    row = DatasetRow(
        id=6,
        vtype="",
        dir_type=1,
        name="obj",
        checksum="",
        parent_id=None,
        last_modified=None,
        anno={},
        etag="",
        version="",
        is_latest=True,
        size=7,
        owner_name="",
        owner_id="",
        path_str="path",
        source="",
        random=1234,
        location=None,
    )
    result = t(None, row)  # pylint: disable=no-value-for-parameter
    assert result == 42


def test_generator():
    @generator(C.name)
    def gen(parent, name):
        yield parent, name

    row = DatasetRow(
        id=6,
        vtype="",
        dir_type=1,
        name="obj",
        checksum="",
        parent_id=None,
        last_modified=None,
        anno={},
        etag="",
        version="",
        is_latest=True,
        size=7,
        owner_name="",
        owner_id="",
        path_str="path",
        source="",
        random=1234,
        location=None,
    )

    assert list(gen(None, row)) == [(row, "obj")]



import io


class FileSlice(io.RawIOBase):
    """A thin wrapper around an existing file object that
    provides a part of its data as an individual file
    object.
    """

    def __init__(self, fileobj, offset, size, name):
        self.fileobj = fileobj
        self.offset = offset
        self.size = size
        self.position = 0
        self.name = name

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return self.fileobj.seekable()

    def tell(self):
        """Return the current file position."""
        return self.position

    def seek(self, position, whence=io.SEEK_SET):
        """Seek to a position in the file."""
        if whence == io.SEEK_SET:
            self.position = min(max(position, 0), self.size)
        elif whence == io.SEEK_CUR:
            if position < 0:
                self.position = max(self.position + position, 0)
            else:
                self.position = min(self.position + position, self.size)
        elif whence == io.SEEK_END:
            self.position = max(min(self.size + position, self.size), 0)
        else:
            raise ValueError("Invalid argument")
        return self.position

    def readinto(self, b):
        max_size = self.size - self.position
        if max_size <= 0:
            return 0
        self.fileobj.seek(self.offset + self.position)
        if len(b) > max_size:
            b = memoryview(b)[:max_size]
        res = self.fileobj.readinto(b)
        if res != len(b):
            raise RuntimeError("unexpected end of data")
        self.position += res
        return res



from jmespath import exceptions
from jmespath.functions import Functions, signature

from dql.jmes_sql.field_ops import FieldOps


class FunctionsSql(Functions):
    def __init__(self, transpiler):
        super().__init__()
        self.transpiler = transpiler

    def call_function(self, function_name, resolved_args):
        try:
            spec = self.FUNCTION_TABLE[function_name]
        except KeyError:
            raise exceptions.UnknownFunctionError(
                "Unknown function: %s()" % function_name
            )
        function = spec["function"]
        self._validate_arguments(
            resolved_args,
            spec["signature"],
            function_name,
        )
        return function(self, *resolved_args)

    @signature({"types": ["array", "string"]}, {"types": []})
    def _func_contains(self, subject, search):
        if not subject:
            return f"list_contains({FieldOps.CUST_FIELD}, '{search}')"

        cmd = FieldOps.get_cmd(subject)
        if cmd:
            value = FieldOps.get_cmd(search)
            if not value:
                value = f"'{search}'"
            return f"list_contains({cmd}, {value})"

        field = FieldOps.resolve_filed(subject)
        if field:
            return field + f" LIKE '%{search}%'"

        return f"'{subject}' LIKE '%{search}%'"



from sqlalchemy import JSON, Column, Float
from sqlalchemy.dialects.sqlite import dialect as sqlite_dialect
from sqlalchemy.schema import CreateTable

from dql.data_storage.schema import DatasetRow


def test_dataset_table_compilation():
    table = DatasetRow.new_table(
        "ds-1",
        custom_columns=[
            Column("score", Float, nullable=False),
            Column("meta_info", JSON),
        ],
    )
    result = CreateTable(table, if_not_exists=True).compile(dialect=sqlite_dialect())

    assert result.string == (
        "\n"
        'CREATE TABLE IF NOT EXISTS "ds-1" (\n'
        "\tid INTEGER NOT NULL, \n"
        "\tvtype TEXT NOT NULL, \n"
        "\tdir_type INTEGER, \n"
        "\tparent_id INTEGER, \n"
        "\tname TEXT NOT NULL, \n"
        "\tchecksum TEXT, \n"
        "\tetag TEXT, \n"
        "\tversion TEXT, \n"
        "\tis_latest BOOLEAN, \n"
        "\tlast_modified DATETIME, \n"
        "\tsize BIGINT NOT NULL, \n"
        "\towner_name TEXT, \n"
        "\towner_id TEXT, \n"
        "\tpath_str TEXT, \n"
        "\tanno JSON, \n"
        "\trandom BIGINT NOT NULL, \n"
        "\tlocation JSON, \n"
        "\tsource TEXT NOT NULL, \n"
        "\tscore FLOAT NOT NULL, \n"
        "\tmeta_info JSON, \n"
        "\tPRIMARY KEY (id)\n"
        ")\n"
        "\n"
    )



import io
import json
import os.path
from random import getrandbits

import pytest
from sqlalchemy import Integer, String

from dql.data_storage.abstract import RANDOM_BITS
from dql.dataset import Status as DatasetStatus
from dql.loader import DataView
from dql.query import C, DatasetQuery, Object, generator, udf
from dql.query.builtins import checksum, index_tar

from ..utils import TARRED_TREE


@pytest.mark.parametrize("from_path", [True, False])
@pytest.mark.parametrize("save", [True, False])
@pytest.mark.parametrize(
    "cloud_type,version_aware",
    [("s3", True)],
    indirect=True,
)
def test_filter(cloud_test_catalog, save, from_path):
    catalog = cloud_test_catalog.catalog
    conf = cloud_test_catalog.client_config
    # ensure we can select a subset of a bucket properly
    path = f"{cloud_test_catalog.src_uri}/cats"
    if from_path:
        ds = DatasetQuery(path=path, catalog=catalog, client_config=conf)
    else:
        sources = [path]
        globs = [s.rstrip("/") + "/*" for s in sources]
        catalog.index(sources, client_config=conf)
        catalog.create_shadow_dataset(
            "animals", globs, client_config=conf, recursive=True
        )
        ds = DatasetQuery(name="animals", catalog=catalog)
    q = ds.filter(C.size < 13).filter(C.path_str.glob("cats/*") | (C.size < 4))
    if save:
        ds_name = "animals_cats"
        q.save(ds_name)
        new_query = DatasetQuery(name=ds_name, catalog=catalog)
        result = new_query.results()
        dataset_record = catalog.get_dataset(ds_name)
        assert dataset_record.status == DatasetStatus.COMPLETE
    else:
        result = q.results()
    assert len(result) == 2


@pytest.mark.parametrize(
    "cloud_type,version_aware",
    [("s3", True)],
    indirect=True,
)
def test_udf(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog
    conf = cloud_test_catalog.client_config
    sources = [cloud_test_catalog.src_uri]
    globs = [s.rstrip("/") + "/*" for s in sources]
    catalog.index(sources, client_config=conf)
    catalog.create_shadow_dataset("animals", globs, client_config=conf, recursive=True)

    @udf(Integer, C.name)
    def name_len(name):
        # A very simple udf.
        return len(name)

    q = (
        DatasetQuery(name="animals", catalog=catalog)
        .filter(C.size < 13)
        .filter(C.path_str.glob("cats/*") | (C.size < 4))
        .add_signal(name_len, "name_len")
    )
    result = q.results()

    assert len(result) == 3


def to_str(buf) -> str:
    return io.TextIOWrapper(buf, encoding="utf8").read()


@pytest.mark.parametrize(
    "cloud_type,version_aware",
    [("s3", True)],
    indirect=True,
)
def test_udf_object_param(cloud_test_catalog, dogs_shadow_dataset):
    # Setup catalog.
    catalog = cloud_test_catalog.catalog
    catalog.client_config = cloud_test_catalog.client_config

    @udf(String, C.name, Object(to_str))
    def signal(name, obj):
        # A very simple udf.
        return name + " -> " + obj

    q = DatasetQuery(name=dogs_shadow_dataset.name, catalog=catalog).add_signal(
        signal, "signal"
    )
    result = q.results()

    assert len(result) == 4
    signals = [r[-1] for r in result]
    assert signals == ["dog1 -> woof", "dog2 -> arf", "dog3 -> bark", "dog4 -> ruff"]


@pytest.mark.parametrize(
    "cloud_type,version_aware",
    [("s3", True)],
    indirect=True,
)
def test_union(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog
    conf = cloud_test_catalog.client_config
    sources = [str(cloud_test_catalog.src_uri)]
    catalog.index(sources, client_config=conf)

    src = cloud_test_catalog.src_uri
    catalog.create_shadow_dataset(
        "dogs", [f"{src}/dogs/*"], client_config=conf, recursive=True
    )
    catalog.create_shadow_dataset(
        "cats", [f"{src}/cats/*"], client_config=conf, recursive=True
    )

    dogs = DatasetQuery(name="dogs", catalog=catalog)
    cats = DatasetQuery(name="cats", catalog=catalog)

    (dogs | cats).save("dogs_cats")

    result = DatasetQuery(name="dogs_cats", catalog=catalog).results()
    assert len(result) == 6


def test_loader_from_query(cloud_test_catalog):
    catalog = cloud_test_catalog.catalog
    conf = cloud_test_catalog.client_config
    src_uri = cloud_test_catalog.src_uri
    catalog.index([src_uri], client_config=conf)
    catalog.create_shadow_dataset(
        "animals", [f"{src_uri}/*"], client_config=conf, recursive=True
    )
    q = DatasetQuery(name="animals", catalog=catalog).filter(C.path_str.glob("cats/*"))

    def transform(row, sample):
        return sample, row.name[-1]

    ds = DataView.from_query(
        q, reader=to_str, transform=transform, catalog=catalog, client_config=conf
    )
    assert set(ds) == {("meow", "1"), ("mrow", "2")}


def create_subobject(row, path_str: str, size: int):
    """Create a subobject with this row as its parent."""
    return {
        # Values that depend on the parent object.
        "vtype": "fake",
        "dir_type": 0,
        "parent_id": row.id,
        # Inherited values.
        "owner_name": row.owner_name,
        "owner_id": row.owner_id,
        "is_latest": row.is_latest,
        "source": row.source,
        # User-provided values.
        "name": os.path.normpath(os.path.basename(path_str)),
        "path_str": _sub_path(row, path_str),
        "size": size,
        # Unspecified.
        "version": None,
        "etag": None,
        # Generated.
        "random": getrandbits(RANDOM_BITS),
    }


def _sub_path(row, child_path: str) -> str:
    """Define the path of the subobject in relation to the parent."""
    # The parent/child annotation logic will need to be updated.
    return f"{row.path_str}:{child_path}"


@pytest.mark.parametrize(
    "cloud_type,version_aware",
    [("s3", True)],
    indirect=True,
)
def test_row_generator(cloud_test_catalog, dogs_shadow_dataset):
    # Setup catalog.
    catalog = cloud_test_catalog.catalog
    catalog.client_config = cloud_test_catalog.client_config

    @generator(C.name, C.path_str)
    def gen(parent, _, __):
        # A very simple generator.
        yield create_subobject(parent, "subobject", 50)

    q = DatasetQuery(name=dogs_shadow_dataset.name, catalog=catalog).generate(gen)
    result = q.results()

    assert len(result) == 8

    signals = {r[13]: r[4] for r in result}  # path_str: name
    assert signals == {
        "dogs/dog1": "dog1",
        "dogs/dog1:subobject": "subobject",
        "dogs/dog2": "dog2",
        "dogs/dog2:subobject": "subobject",
        "dogs/dog3": "dog3",
        "dogs/dog3:subobject": "subobject",
        "dogs/others/dog4": "dog4",
        "dogs/others/dog4:subobject": "subobject",
    }


@pytest.mark.parametrize("tree", [TARRED_TREE], indirect=True)
def test_index_tar(cloud_test_catalog):
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    catalog.client_config = ctc.client_config
    catalog.index([ctc.src_uri])
    catalog.create_shadow_dataset("animals", [ctc.src_uri])

    q = DatasetQuery(name="animals", catalog=catalog).generate(index_tar)
    q.save("extracted")

    rows = list(catalog.ls_dataset_rows("extracted"))
    assert {row.name for row in rows} == {
        "animals.tar",
        "cat1",
        "cat2",
        "description",
        "dog1",
        "dog2",
        "dog3",
        "dog4",
    }

    offsets = [
        json.loads(row.location)[0]["offset"]
        for row in rows
        if row.name != "animals.tar"
    ]
    # Check that offsets are unique integers
    assert all(isinstance(offset, int) for offset in offsets)
    assert len(set(offsets)) == len(offsets)

    assert all(row.vtype == "tar" for row in rows if row.name != "animals.tar")


@pytest.mark.parametrize(
    "cloud_type,version_aware",
    [("s3", True)],
    indirect=True,
)
def test_checksum_udf(cloud_test_catalog, dogs_shadow_dataset):
    # Setup catalog.
    catalog = cloud_test_catalog.catalog
    catalog.client_config = cloud_test_catalog.client_config

    q = DatasetQuery(name=dogs_shadow_dataset.name, catalog=catalog).add_signal(
        checksum
    )
    result = q.results()

    assert len(result) == 4


@pytest.mark.parametrize("tree", [TARRED_TREE], indirect=True)
def test_tar_loader(cloud_test_catalog):
    ctc = cloud_test_catalog
    catalog = ctc.catalog
    catalog.client_config = ctc.client_config
    catalog.index([ctc.src_uri])
    catalog.create_shadow_dataset("animals", [ctc.src_uri])
    q = DatasetQuery(name="animals", catalog=catalog).generate(index_tar)
    q.save("extracted")

    q = DatasetQuery(name="extracted", catalog=catalog).filter(
        C.path_str.glob("*/cats/*")
    )
    assert len(q.results()) == 2

    def transform(row, sample):
        return sample, row.name[-1]

    ds = DataView.from_query(q, reader=to_str, transform=transform, catalog=catalog)
    assert set(ds) == {("meow", "1"), ("mrow", "2")}



from typing import TYPE_CHECKING, Generic, List, Optional, Sequence, Type, TypeVar

import sqlalchemy as sa
from sqlalchemy import (
    JSON,
    BigInteger,
    Boolean,
    Column,
    DateTime,
    Integer,
    MetaData,
    Text,
)
from sqlalchemy.sql.expression import true

from dql.node import DirTypeGroup

if TYPE_CHECKING:
    from sqlalchemy import Engine
    from sqlalchemy.sql.base import Executable
    from sqlalchemy.sql.selectable import Select


class Table:
    def __init__(self, name: str, metadata: Optional["MetaData"] = None):
        self.metadata: "MetaData" = metadata if metadata is not None else MetaData()
        self.name: str = name
        self.table: "sa.Table" = self.get_table()

    @property
    def columns(self) -> List[Column]:
        return self.table.columns

    @property
    def c(self):
        return self.columns

    @classmethod
    def default_columns(cls) -> List[Column]:
        return []

    @property
    def custom_columns(self):
        """List of custom columns added to the table."""
        default_cols = [c.name for c in self.default_columns()]
        return [c for c in self.columns if c.name not in default_cols]

    def get_table(self) -> "sa.Table":
        table = self.metadata.tables.get(self.name)
        if table is None:
            table = sa.Table(
                self.name,
                self.metadata,
                *self.default_columns(),
            )
        return table

    def apply_conditions(self, query: "Executable") -> "Executable":
        """
        Apply any conditions that belong on all selecting queries.

        This could be used to filter tables that use access control.
        """
        return query

    def select(self, *columns):
        if not columns:
            query = self.table.select()
        else:
            query = sa.select(*columns).select_from(self.table)
        return self.apply_conditions(query)

    def insert(self):
        return self.table.insert()

    def update(self):
        return self.apply_conditions(self.table.update())

    def delete(self):
        return self.apply_conditions(self.table.delete())


class BaseDatasetRow(Table):
    @classmethod
    def default_columns(cls) -> List[Column]:
        return [
            Column("id", Integer, primary_key=True),
            Column("vtype", Text, nullable=False, index=True),
            Column("dir_type", Integer, index=True),
            Column("parent_id", Integer, index=True),
            Column("name", Text, nullable=False, index=True),
            Column("checksum", Text),
            Column("etag", Text),
            Column("version", Text),
            Column("is_latest", Boolean),
            Column("last_modified", DateTime),
            Column("size", BigInteger, nullable=False, index=True),
            Column("owner_name", Text),
            Column("owner_id", Text),
            Column("path_str", Text, index=True),
            Column("anno", JSON),
            Column("random", BigInteger, nullable=False),
            Column("location", JSON),
        ]

    @staticmethod
    def copy_signal_column(column: Column):
        """
        Copy a sqlalchemy Column object intended for use as a signal column.

        This does not copy all attributes as certain attributes such as
        table are too context-dependent and the purpose of this function is
        adding a signal column from one table to another table.

        We can't use Column.copy() as it only works in certain contexts.
        See https://github.com/sqlalchemy/sqlalchemy/issues/5953
        """
        return Column(
            column.name,
            column.type,
            primary_key=column.primary_key,
            index=column.index,
            nullable=column.nullable,
            default=column.default,
            server_default=column.server_default,
            unique=column.unique,
        )

    @classmethod
    def new_table(
        cls,
        name: str,
        custom_columns: Sequence["Column"] = (),
        metadata: Optional["MetaData"] = None,
    ):
        # copy columns, since re-using the same objects from another table
        # may raise an error
        custom_columns = [cls.copy_signal_column(c) for c in custom_columns]
        if metadata is None:
            metadata = MetaData()
        return sa.Table(
            name,
            metadata,
            *cls.default_columns(),
            *custom_columns,
        )


class Node(BaseDatasetRow):
    def __init__(self, name: str, source: str, metadata: Optional["MetaData"] = None):
        super().__init__(name, metadata=metadata)
        self.source = source

    @classmethod
    def default_columns(cls) -> List[Column]:
        return [
            *super().default_columns(),
            Column("partial_id", Integer),
            Column("valid", Boolean, default=True, nullable=False),
        ]

    def dataset_query(
        self,
        *column_names: str,
    ) -> "Select":
        if not column_names:
            column_names = DATASET_CORE_COLUMN_NAMES
        column_objects = [
            sa.literal(self.source).label("source") if c == "source" else self.c[c]
            for c in column_names
        ]
        # include all object types - file, tar archive, tar file (subobject)
        return self.select(*column_objects).where(
            self.c.dir_type.in_(DirTypeGroup.FILE)
            & (self.c.valid == true())
            & (self.c.is_latest == true())
        )


class DatasetRow(BaseDatasetRow):
    def __init__(
        self, name: str, engine: "Engine", metadata: Optional["MetaData"] = None
    ):
        self.engine = engine
        super().__init__(name, metadata)

    def get_table(self) -> "sa.Table":
        table = self.metadata.tables.get(self.name)
        if table is None:
            return sa.Table(
                self.name,
                self.metadata,
                extend_existing=True,
                autoload_with=self.engine,
            )
        return table

    @classmethod
    def default_columns(cls) -> List[Column]:
        return [
            *super().default_columns(),
            Column("source", Text, nullable=False),
        ]


DATASET_CORE_COLUMN_NAMES = tuple([col.name for col in DatasetRow.default_columns()])

NodeT = TypeVar("NodeT", bound=Node)
DatasetRowT = TypeVar("DatasetRowT", bound=DatasetRow)


class Schema(Generic[NodeT, DatasetRowT]):
    node_cls: Type[NodeT]
    dataset_row_cls: Type[DatasetRowT]


class DefaultSchema(Schema[Node, DatasetRow]):
    def __init__(self):
        self.node_cls = Node
        self.dataset_row_cls = DatasetRow
